@@ -19,6 +19,7 @@ build_tests=false
 build_verbose=false
 clean_build=true
 collective_trace=true
+enable_code_coverage=false
 enable_ninja=""
 install_dependencies=false
 install_library=false
@@ -26,6 +27,7 @@ install_prefix="${ROCM_PATH}"
 log_trace=false
 msccl_kernel_enabled=true
 mscclpp_enabled=true
+enable_mscclpp_clip=false
 num_parallel_jobs=$(nproc)
 npkit_enabled=false
 openmp_test_enabled=false
@@ -33,6 +35,7 @@ roctx_enabled=true
 run_tests=false
 run_tests_all=false
 time_trace=false
+force_reduce_pipeline=false
 
 # #################################################
 # helper functions
@@ -42,12 +45,14 @@ function display_help()
     echo "RCCL build & installation helper script"
     echo " Options:"
     echo "       --address-sanitizer     Build with address sanitizer enabled"
+    echo "    -c|--enable-code-coverage  Enable Code Coverage"
     echo "    -d|--dependencies          Install RCCL dependencies"
     echo "       --debug                 Build debug library"
     echo "       --enable_backtrace      Build with custom backtrace support"
     echo "       --disable-colltrace     Build without collective trace"
     echo "       --disable-msccl-kernel  Build without MSCCL kernels"
     echo "       --disable-mscclpp       Build without MSCCL++ support"
+    echo "       --enable-mscclpp-clip   Build MSCCL++ with clip wrapper on bfloat16 and half addition routines"
     echo "       --disable-roctx         Build without ROCTX logging"
     echo "    -f|--fast                  Quick-build RCCL (local gpu arch only, no backtrace, and collective trace support)"
     echo "    -h|--help                  Prints this help message"
@@ -61,13 +66,13 @@ function display_help()
     echo "       --openmp-test-enable    Enable OpenMP in rccl unit tests"
     echo "    -p|--package_build         Build RCCL package"
     echo "       --prefix                Specify custom directory to install RCCL to (default: \`/opt/rocm\`)"
-    echo "       --rm-legacy-include-dir Remove legacy include dir Packaging added for file/folder reorg backward compatibility"
     echo "       --run_tests_all         Run all rccl unit tests (must be built already)"
     echo "    -r|--run_tests_quick       Run small subset of rccl unit tests (must be built already)"
     echo "       --static                Build RCCL as a static library instead of shared library"
     echo "    -t|--tests_build           Build rccl unit tests, but do not run"
     echo "       --time-trace            Plot the build time of RCCL (requires \`ninja-build\` package installed on the system)"
     echo "       --verbose               Show compile commands"
+    echo "       --force-reduce-pipeline Force reduce_copy sw pipeline to be used for every reduce-based collectives and datatypes"
 }
 
 # #################################################
@@ -77,7 +82,7 @@ function display_help()
 # check if we have a modern version of getopt that can handle whitespace and long parameters
 getopt -T
 if [[ "$?" -eq 4 ]]; then
-    GETOPT_PARSE=$(getopt --name "${0}" --options dfhij:lprt --longoptions address-sanitizer,dependencies,debug,enable_backtrace,disable-colltrace,disable-msccl-kernel,disable-mscclpp,fast,help,install,jobs:,local_gpu_only,amdgpu_targets:,no_clean,npkit-enable,log-trace,openmp-test-enable,roctx-enable,package_build,prefix:,rm-legacy-include-dir,run_tests_all,run_tests_quick,static,tests_build,time-trace,verbose -- "$@")
+    GETOPT_PARSE=$(getopt --name "${0}" --options cdfhij:lprt --longoptions address-sanitizer,dependencies,debug,enable-code-coverage,enable_backtrace,disable-colltrace,disable-msccl-kernel,disable-mscclpp,fast,help,install,jobs:,local_gpu_only,amdgpu_targets:,no_clean,npkit-enable,log-trace,openmp-test-enable,roctx-enable,package_build,prefix:,rm-legacy-include-dir,run_tests_all,run_tests_quick,static,tests_build,time-trace,force-reduce-pipeline,verbose -- "$@")
 else
     echo "Need a new version of getopt"
     exit 1
@@ -93,12 +98,14 @@ eval set -- "${GETOPT_PARSE}"
 while true; do
     case "${1}" in
          --address-sanitizer)        build_address_sanitizer=true;                                                                     shift ;;
+    -c | --enable-code-coverage)     enable_code_coverage=true;                                                                        shift ;;
     -d | --dependencies)             install_dependencies=true;                                                                        shift ;;
          --debug)                    build_release=false;                                                                              shift ;;
          --enable_backtrace)         build_bfd=true;                                                                                   shift ;;
          --disable-colltrace)        collective_trace=false;                                                                           shift ;;
          --disable-msccl-kernel)     msccl_kernel_enabled=false;                                                                       shift ;;
          --disable-mscclpp)          mscclpp_enabled=false;                                                                            shift ;;
+         --enable-mscclpp-clip)      enable_mscclpp_clip=true;                                                                         shift ;;
          --disable-roctx)            roctx_enabled=false;                                                                              shift ;;
     -f | --fast)                     build_local_gpu_only=true; collective_trace=false; msccl_kernel_enabled=false;                    shift ;;
     -h | --help)                     display_help;                                                                                     exit 0 ;;
@@ -112,13 +119,13 @@ while true; do
          --openmp-test-enable)       openmp_test_enabled=true;                                                                         shift ;;
     -p | --package_build)            build_package=true;                                                                               shift ;;
          --prefix)                   install_library=true; install_prefix=${2};                                                        shift 2 ;;
-         --rm-legacy-include-dir)    build_freorg_bkwdcomp=false;                                                                      shift ;;
     -r | --run_tests_quick)          run_tests=true;                                                                                   shift ;;
          --run_tests_all)            run_tests=true; run_tests_all=true;                                                               shift ;;
          --static)                   build_static=true;                                                                                shift ;;
     -t | --tests_build)              build_tests=true;                                                                                 shift ;;
          --time-trace)               time_trace=true;                                                                                  shift ;;
          --verbose)                  build_verbose=true;                                                                               shift ;;
+         --force-reduce-pipeline)    force_reduce_pipeline=true;                                                                        shift ;;
     --) shift ; break ;;
     *)  echo "Unexpected command line parameter received; aborting";
         exit 1
@@ -205,16 +212,16 @@ if [[ "${build_address_sanitizer}" == true ]]; then
     cmake_common_options="${cmake_common_options} -DBUILD_ADDRESS_SANITIZER=ON"
 fi
 
+# Enable code coverage
+if [[ "${enable_code_coverage}" == true ]]; then
+    cmake_common_options="${cmake_common_options} -DENABLE_CODE_COVERAGE=ON"
+fi
+
 # Backtrace support
 if [[ "${build_bfd}" == true ]]; then
     cmake_common_options="${cmake_common_options} -DBUILD_BFD=ON"
 fi
 
-# Backward compatibility wrappers
-if [[ "${build_freorg_bkwdcomp}" == true ]]; then
-    cmake_common_options="${cmake_common_options} -DBUILD_FILE_REORG_BACKWARD_COMPATIBILITY=ON"
-fi
-
 # Build local GPU arch only
 if [[ "${build_local_gpu_only}" == true ]]; then
     cmake_common_options="${cmake_common_options} -DBUILD_LOCAL_GPU_TARGET_ONLY=ON"
@@ -244,6 +251,10 @@ if [[ "${mscclpp_enabled}" == false ]]; then
     cmake_common_options="${cmake_common_options} -DENABLE_MSCCLPP=OFF"
 fi
 
+if [[ "${enable_mscclpp_clip}" == true ]]; then
+    cmake_common_options="${cmake_common_options} -DENABLE_MSCCLPP_CLIP=ON"
+fi
+
 # Install dependencies
 if [[ "${install_dependencies}" == true ]]; then
     cmake_common_options="${cmake_common_options} -DINSTALL_DEPENDENCIES=ON"
@@ -269,115 +280,14 @@ if [[ "${openmp_test_enabled}" == true ]]; then
     cmake_common_options="${cmake_common_options} -DOPENMP_TESTS_ENABLED=ON"
 fi
 
+# Force Reduce pipeline
+if [[ "${force_reduce_pipeline}" == true ]]; then
+    cmake_common_options="${cmake_common_options} -DFORCE_REDUCE_PIPELINING=ON"
+fi
+
 # Enable NPKit
-npkit_options=""
 if [[ "${npkit_enabled}" == true ]]; then
-    npkit_options="-DENABLE_NPKIT \
-    -DENABLE_NPKIT_EVENT_TIME_SYNC_GPU \
-    -DENABLE_NPKIT_EVENT_TIME_SYNC_CPU \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_UPDOWN_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_UPDOWN_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_SPLIT_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_SPLIT_EXIT \
-    -DENABLE_NPKIT_EVENT_COPY_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_COPY_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_DIRECT_COPY_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_DIRECT_COPY_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_DIRECT_RECV_ENTRY \
-    -DENABLE_NPKIT_EVENT_DIRECT_RECV_EXIT \
-    -DENABLE_NPKIT_EVENT_DIRECT_RECV_COPY_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_DIRECT_RECV_COPY_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_DIRECT_RECV_REDUCE_COPY_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_DIRECT_RECV_REDUCE_COPY_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_DIRECT_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_DIRECT_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_DIRECT_SEND_FROM_OUTPUT_ENTRY \
-    -DENABLE_NPKIT_EVENT_DIRECT_SEND_FROM_OUTPUT_EXIT \
-    -DENABLE_NPKIT_EVENT_RECV_ENTRY \
-    -DENABLE_NPKIT_EVENT_RECV_EXIT \
-    -DENABLE_NPKIT_EVENT_RECV_COPY_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_RECV_COPY_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_RECV_REDUCE_COPY_ENTRY \
-    -DENABLE_NPKIT_EVENT_RECV_REDUCE_COPY_EXIT \
-    -DENABLE_NPKIT_EVENT_RECV_REDUCE_COPY_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_RECV_REDUCE_COPY_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_RECV_REDUCE_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_RECV_REDUCE_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_SEND_FROM_OUTPUT_ENTRY \
-    -DENABLE_NPKIT_EVENT_SEND_FROM_OUTPUT_EXIT \
-    -DENABLE_NPKIT_EVENT_PRIM_SIMPLE_WAIT_PEER_ENTRY \
-    -DENABLE_NPKIT_EVENT_PRIM_SIMPLE_WAIT_PEER_EXIT \
-    -DENABLE_NPKIT_EVENT_PRIM_SIMPLE_REDUCE_OR_COPY_MULTI_ENTRY \
-    -DENABLE_NPKIT_EVENT_PRIM_SIMPLE_REDUCE_OR_COPY_MULTI_EXIT \
-    -DENABLE_NPKIT_EVENT_PRIM_LL_WAIT_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_PRIM_LL_WAIT_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_PRIM_LL_DATA_PROCESS_ENTRY \
-    -DENABLE_NPKIT_EVENT_PRIM_LL_DATA_PROCESS_EXIT \
-    -DENABLE_NPKIT_EVENT_PRIM_LL128_WAIT_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_PRIM_LL128_WAIT_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_PRIM_LL128_DATA_PROCESS_ENTRY \
-    -DENABLE_NPKIT_EVENT_PRIM_LL128_DATA_PROCESS_EXIT \
-    -DENABLE_NPKIT_EVENT_NET_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_NET_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_NET_TEST_ENTRY \
-    -DENABLE_NPKIT_EVENT_NET_TEST_EXIT \
-    -DENABLE_NPKIT_EVENT_NET_RECV_ENTRY \
-    -DENABLE_NPKIT_EVENT_NET_RECV_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_RECV_REDUCE_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_RECV_REDUCE_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_DIRECT_RECV_REDUCE_COPY_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_DIRECT_RECV_REDUCE_COPY_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_DIRECT_RECV_COPY_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_DIRECT_RECV_COPY_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_DIRECT_RECV_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_DIRECT_RECV_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_UPDOWN_REDUCE_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_UPDOWN_REDUCE_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_UPDOWN_BROADCAST_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_UPDOWN_BROADCAST_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_SPLIT_REDUCE_BROADCAST_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_SPLIT_REDUCE_BROADCAST_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_SPLIT_REDUCE_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_SPLIT_REDUCE_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_SPLIT_BROADCAST_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_REDUCE_TREE_SPLIT_BROADCAST_EXIT \
-    -DENABLE_NPKIT_EVENT_SEND_RECV_LOCAL_COPY_ENTRY \
-    -DENABLE_NPKIT_EVENT_SEND_RECV_LOCAL_COPY_EXIT \
-    -DENABLE_NPKIT_EVENT_SEND_RECV_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_SEND_RECV_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_SEND_RECV_RECV_ENTRY \
-    -DENABLE_NPKIT_EVENT_SEND_RECV_RECV_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_RECV_COPY_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_RECV_COPY_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_DIRECT_RECV_ENTRY \
-    -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_DIRECT_RECV_EXIT \
-    -DENABLE_NPKIT_EVENT_MSCCL_GENERIC_OP_ENTRY \
-    -DENABLE_NPKIT_EVENT_MSCCL_GENERIC_OP_EXIT \
-    -DENABLE_NPKIT_EVENT_MSCCL_REDUCE_ENTRY \
-    -DENABLE_NPKIT_EVENT_MSCCL_REDUCE_EXIT \
-    -DENABLE_NPKIT_EVENT_MSCCL_SEND_ENTRY \
-    -DENABLE_NPKIT_EVENT_MSCCL_SEND_EXIT \
-    -DENABLE_NPKIT_EVENT_MSCCL_RECV_ENTRY \
-    -DENABLE_NPKIT_EVENT_MSCCL_RECV_EXIT \
-    -DENABLE_NPKIT_EVENT_MSCCL_RUN_ENTRY \
-    -DENABLE_NPKIT_EVENT_MSCCL_RUN_EXIT \
-    -DENABLE_NPKIT_EVENT_MSCCL_RECV_REDUCE_COPY_ENTRY \
-    -DENABLE_NPKIT_EVENT_MSCCL_RECV_REDUCE_COPY_EXIT \
-    -DENABLE_NPKIT_EVENT_MSCCL_INIT_ENTRY \
-    -DENABLE_NPKIT_EVENT_MSCCL_INIT_EXIT \
-    -DENABLE_NPKIT_EVENT_BROADCAST_RING_ENTRY \
-    -DENABLE_NPKIT_EVENT_BROADCAST_RING_EXIT \
-    -DENABLE_NPKIT_PRIM_COLLECT_DATA_PROCESS_TIME"
+    cmake_common_options="${cmake_common_options} -DENABLE_NPKIT=ON"
 fi
 
 check_exit_code "$?"
@@ -403,10 +313,13 @@ if [[ "${build_tests}" == true ]] || ([[ "${run_tests}" == true ]] && [[ ! -x ./
     cmake_common_options="${cmake_common_options} -DBUILD_TESTS=ON"
 fi
 
+# Add build directory to RPATH for packaging dependency resolution
+cmake_common_options="${cmake_common_options} -DCMAKE_EXE_LINKER_FLAGS=\"-Wl,-rpath,${PWD}\""
+
 # Initiate RCCL CMake
-# Passing NPKIT_FLAGS separately (not as part of ${cmake_common_options}) as
-# ${npkit_options} need to be passed "as-is" i.e. with `-D` to CMakeLists.txt
-${cmake_executable} ${cmake_common_options} -DNPKIT_FLAGS="${npkit_options}" -DONLY_FUNCS="${ONLY_FUNCS}" ../../.
+# Passing ONLY_FUNCS separately (not as part of ${cmake_common_options}) as
+# ${ONLY_FUNCS} is a debug-only feature
+${cmake_executable} ${cmake_common_options} -DONLY_FUNCS="${ONLY_FUNCS}" ../../.
 check_exit_code "$?"
 
 # Enable verbose output from Makefile
