@@ -10,7 +10,7 @@
 
 namespace {
   template<typename T, typename RedOp, typename Proto>
-#if defined(USE_INDIRECT_FUNCTION_CALL) && !defined(__gfx940__) && !defined(__gfx941__) && !defined(__gfx942__)
+#if defined(USE_INDIRECT_FUNCTION_CALL) && !defined(__gfx942__) && !defined(__gfx950__)
   __device__ void runRing(int tid, int nthreads, struct ncclDevWorkColl* work) {
 #else
   __device__ __attribute__((noinline)) void runRing(int tid, int nthreads, struct ncclDevWorkColl* work) {
@@ -67,7 +67,7 @@ namespace {
           }
 
           // final step: recv
-          prims.directRecv(recv_offset + prims_offset, prims_nelem);
+          prims.directRecv(recv_offset + prims_offset, recv_offset + prims_offset, prims_nelem);
         }
       }
     }
