@@ -31,35 +31,16 @@
     ret func(args)
 #endif // end PROFAPI
 
-static __inline__ int ncclTypeSize(ncclDataType_t type) {
-  switch (type) {
-    case ncclInt8:
-    case ncclUint8:
-      return 1;
-    case ncclFloat16:
-#if defined(RCCL_BFLOAT16)
-    case ncclBfloat16:
-#endif
-      return 2;
-    case ncclInt32:
-    case ncclUint32:
-    case ncclFloat32:
-      return 4;
-    case ncclInt64:
-    case ncclUint64:
-    case ncclFloat64:
-      return 8;
-    default:
-      return -1;
-  }
-}
-
 #include "debug.h"
 #include "checks.h"
 #include "rocmwrap.h"
 #include "alloc.h"
 #include "utils.h"
 #include "param.h"
+#ifdef NVTX_NO_IMPL
 #include "nvtx_stub.h"
+#else
+#include "nvtx.h"
+#endif
 
 #endif // end include guard
