@@ -27,4 +27,12 @@ THE SOFTWARE.
 
 RCCL_PARAM_DECLARE(EnableHipGraph);  // Opt-in environment variable for enabling hipGraph
 
+#ifdef RCCL_EXPOSE_STATIC
+#define rccl_static
+#define rccl_static_inline
+#else
+#define rccl_static static
+#define rccl_static_inline static inline
+#endif
+
 #endif
